# Cassandra Read-Side support

This page is specifically about Lagom's support for Cassandra read-sides.  Before reading this, you should familiarize yourself with Lagom's general [[read-side support|ReadSide]].

## Query the Read-Side Database

Let us first look at how a service implementation can retrieve data from Cassandra.

@[imports](code/docs/home/persistence/CassandraReadSideQuery.java)
@[service-impl](code/docs/home/persistence/CassandraReadSideQuery.java)

Note that the [CassandraSession](api/index.html?com/lightbend/lagom/javadsl/persistence/cassandra/CassandraSession.html) is injected in the constructor. `CassandraSession` provides several methods in different flavors for executing queries. The one used in the above example returns a `Source`, i.e. a streamed response. There are also methods for retrieving a list of rows, which can be useful when you know that the result set is small, e.g. when you have included a `LIMIT` clause.

All methods in `CassandraSession` are non-blocking and they return a `CompletionStage` or a `Source`. The statements are expressed in [Cassandra Query Language](https://docs.datastax.com/en/archived/cql/3.3/cql/cqlIntro.html) (CQL) syntax. See [Querying tables](https://docs.datastax.com/en/archived/cql/3.3/cql/cql_using/useQueryDataTOC.html) for information about CQL queries.

## Update the Read-Side

We need to transform the events generated by the [[Persistent Entities|PersistentEntity]] into database tables that can be queried as illustrated in the previous section. For that we will implement a [ReadSideProcessor](api/index.html?com/lightbend/lagom/javadsl/persistence/ReadSideProcessor.html) with assistance from the [CassandraReadSide](api/index.html?com/lightbend/lagom/javadsl/persistence/cassandra/CassandraReadSide.html) support component. It will consume events produced by persistent entities and update one or more tables in Cassandra that are optimized for queries.

This is how a [ReadSideProcessor](api/index.html?com/lightbend/lagom/javadsl/persistence/ReadSideProcessor.html) class looks like before filling in the implementation details:

@[imports](code/docs/home/persistence/CassandraBlogEventProcessor.java)
@[initial](code/docs/home/persistence/CassandraBlogEventProcessor.java)

You can see that we have injected the Cassandra session and Cassandra read-side support, these will be needed later.

You should already have implemented tagging for your events as described in the [[Read-Side documentation|ReadSide]], so first we'll implement the `aggregateTags` method in our read-side processor stub, like so:

@[tag](code/docs/home/persistence/CassandraBlogEventProcessor.java)

### Building the read-side handler

The other method on the `ReadSideProcessor` is `buildHandler`.  This is responsible for creating the [ReadSideHandler](api/index.html?com/lightbend/lagom/javadsl/persistence/ReadSideProcessor.ReadSideHandler.html) that will handle events.  It also gives the opportunity to run two callbacks, one is a global prepare callback, the other is a regular prepare callback.

[CassandraReadSide](api/index.html?com/lightbend/lagom/javadsl/persistence/cassandra/CassandraReadSide.html) has a `builder` method for creating a builder for these handlers, this builder will create a handler that will automatically handle readside offsets for you.  It can be created like so:

@[create-builder](code/docs/home/persistence/CassandraBlogEventProcessor.java)

The argument passed to this method is the ID of the event processor that Lagom will use when it persists offsets to its offset store. The offset store is a Cassandra table, which will be created for you if it doesn't exist. You can manually create this table yourself if you wish, the DDL for its creation is as follows:

```sql
CREATE TABLE IF NOT EXISTS offsetStore (
    eventProcessorId text,
    tag text,
    timeUuidOffset timeuuid,
    sequenceOffset bigint,
    PRIMARY KEY (eventProcessorId, tag)
)
```

### Global prepare

The global prepare callback runs at least once across the whole cluster.  It is intended for doing things like creating tables and preparing any data that needs to be available before read side processing starts.  Read side processors may be sharded across many nodes, and so tasks like creating tables should usually only be done from one node.

The global prepare callback is run from an Akka cluster singleton.  It may be run multiple times - every time a new node becomes the new singleton, the callback will be run.  Consequently, the task must be idempotent.  If it fails, it will be run again using an exponential backoff, and the read side processing of the whole cluster will not start until it has run successfully.

Of course, setting a global prepare callback is completely optional, you may prefer to manage Cassandra tables manually, but it is very convenient for development and test environments to use this callback to create them for you.

Below is an example method that we've implemented to create tables:

@[create-table](code/docs/home/persistence/CassandraBlogEventProcessor.java)

It can then be registered as the global prepare callback in the `buildHandler` method:

@[register-global-prepare](code/docs/home/persistence/CassandraBlogEventProcessor.java)

### Prepare

In addition to the global prepare callback, there is also a prepare callback. This will be executed once per shard, when the read side processor starts up.  It can be used for preparing statements in order to optimize Cassandra's handling of them.

Again this callback is optional, here is an example of how to prepare a statement for updating the table:

@[prepare-statements](code/docs/home/persistence/CassandraBlogEventProcessor.java)

And then to register them:

@[register-prepare](code/docs/home/persistence/CassandraBlogEventProcessor.java)

### Registering your read-side processor

Once you've created your read-side processor, you need to register it with Lagom. This is done using the [`ReadSide`](api/index.html?com/lightbend/lagom/javadsl/persistence/ReadSide.html) component:

@[register-event-processor](code/docs/home/persistence/BlogServiceImpl3.java)

### Event handlers

The event handlers take an event, and return a list of bound statements.  Rather than executing updates in the handler itself, it is recommended that you return the statements that you want to execute to Lagom.  This allows Lagom to batch those statements with the offset table update statement, which Lagom will then executed as a logged batch, which Cassandra executes atomically.  By doing this you can ensure exactly once processing of all events, otherwise processing may be at least once.

Here's an example callback for handling the `PostAdded` event:

@[post-added](code/docs/home/persistence/CassandraBlogEventProcessor.java)

This can then be registered with the builder using `setEventHandler`:

@[set-event-handler](code/docs/home/persistence/CassandraBlogEventProcessor.java)

Once you have finished registering all your event handlers, you can invoke the `build` method and return the built handler:

@[build](code/docs/home/persistence/CassandraBlogEventProcessor.java)

## Underlying implementation

The `CassandraSession` is using the [Datastax Java Driver for Apache Cassandra](https://github.com/datastax/java-driver).

Each `ReadSideProcessor` instance is executed by an [Actor](https://doc.akka.io/docs/akka/2.6/actors.html?language=java) that is managed by [Akka Cluster Sharding](https://doc.akka.io/docs/akka/2.6/cluster-sharding.html?language=java). The processor consumes a stream of persistent events delivered by the `eventsByTag` [Persistence Query](https://doc.akka.io/docs/akka/2.6/persistence-query.html?language=java) implemented by [akka-persistence-cassandra](https://doc.akka.io/docs/akka-persistence-cassandra/0.99/). The tag corresponds to the `tag` defined by the `AggregateEventTag`.

